<?php
/**
 * Zend Framework (http://framework.zend.com/)
 *
 * @link      http://github.com/zendframework/zf2 for the canonical source repository
 * @copyright Copyright (c) 2005-2013 Zend Technologies USA Inc. (http://www.zend.com)
 * @license   http://framework.zend.com/license/new-bsd New BSD License
 */
require_once CLASS_DIR . '/RouteMatch.php';
/**
 * Segment route.
 */
class RouteSegment {
    /**
     * Cache for the encode output.
     *
     * @var array
     */
    protected static $cacheEncode = array();

    /**
     * Map of allowed special chars in path segments.
     *
     * http://tools.ietf.org/html/rfc3986#appendix-A
     * segement      = *pchar
     * pchar         = unreserved / pct-encoded / sub-delims / ":" / "@"
     * unreserved    = ALPHA / DIGIT / "-" / "." / "_" / "~"
     * sub-delims    = "!" / "$" / "&" / "'" / "(" / ")"
     *               / "*" / "+" / "," / ";" / "="
     *
     * @var array
     */
    protected static $urlencodeCorrectionMap = array(
        '%21' => "!", // sub-delims
        '%24' => "$", // sub-delims
        '%26' => "&", // sub-delims
        '%27' => "'", // sub-delims
        '%28' => "(", // sub-delims
        '%29' => ")", // sub-delims
        '%2A' => "*", // sub-delims
        '%2B' => "+", // sub-delims
        '%2C' => ",", // sub-delims
//      '%2D' => "-", // unreserved - not touched by rawurlencode
//      '%2E' => ".", // unreserved - not touched by rawurlencode
        '%3A' => ":", // pchar
        '%3B' => ";", // sub-delims
        '%3D' => "=", // sub-delims
        '%40' => "@", // pchar
//      '%5F' => "_", // unreserved - not touched by rawurlencode
//      '%7E' => "~", // unreserved - not touched by rawurlencode
    );

    /**
     * String of the route.
     *
     * @var string
     */
    protected $route;

    /**
     * Parts of the route.
     *
     * @var array
     */
    protected $parts;

    /**
     * Regex used for matching the route.
     *
     * @var string
     */
    protected $regex;

    /**
     * Map from regex groups to parameter names.
     *
     * @var array
     */
    protected $paramMap = array();

    /**
     * Tablica transformacji dla parametrw przekazywanych do url'a i odbieranych z
     * adresu URL
     *
     * @var array
     */
    protected $trans = array();

    /**
     * Params of the route.
     *
     * @var array
     */
    protected $params;

    /**
     * Default values.
     *
     * @var array
     */
    protected $defaults;

    /**
     * List of assembled parameters.
     *
     * @var array
     */
    protected $assembledParams = array();

    /**
     * Create a new regex route.
     *
     * @param  string $route
     * @param  array  $params
     * @param  array  $defaults
     */
    public function __construct($route, array $params = array(), array $defaults = array(), array $trans = array())
    {
        $this->route    = $route;
        $this->defaults = $defaults;
        $this->params   = $params;
        $this->trans    = $trans;
        $this->parts    = $this->parseRouteDefinition($route);
        $this->regex    = $this->buildRegex($this->parts, $params);
    }

    /**
     * factory(): defined by RouteInterface interface.
     *
     * @see    \Zend\Mvc\Router\RouteInterface::factory()
     * @param  array|Traversable $options
     * @return Segment
     * @throws Exception\InvalidArgumentException
     */
    public static function factory($options = array())
    {
        if ($options instanceof Traversable) {
            $options = ArrayUtils::iteratorToArray($options);
        } elseif (!is_array($options)) {
            throw new Exception(__METHOD__ . ' expects an array or Traversable set of options');
        }

        if (!isset($options['route'])) {
            throw new Exception('Missing "route" in options array');
        }

        if (!isset($options['params'])) {
            $options['params'] = array();
        }

        if (!isset($options['defaults'])) {
            $options['defaults'] = array();
        }

        if (!isset($options['trans']) || !is_array($options['trans']) ) {
            $options['trans'] = array();
        }

        return new RouteSegment($options['route'], $options['params'], $options['defaults'], $options['trans']);
    }

    /**
     * Parse a route definition.
     *
     * @param  string $def
     * @return array
     * @throws Exception\RuntimeException
     */
    protected function parseRouteDefinition($def)
    {
        $currentPos = 0;
        $length     = strlen($def);
        $parts      = array();
        $levelParts = array(&$parts);
        $level      = 0;

        while ($currentPos < $length) {
            preg_match('(\G(?P<literal>[^:{\[\]]*)(?P<token>[:{\[\]]|$))', $def, $matches, 0, $currentPos);

            $currentPos += strlen($matches[0]);

            if (!empty($matches['literal'])) {
                $levelParts[$level][] = array('literal', $matches['literal']);
            }

            if ($matches['token'] === ':') {
                if (!preg_match('(\G(?P<name>[^:/{\.\[\]]+)(?:{(?P<delimiters>[^}]+)})?:?)', $def, $matches, 0, $currentPos)) {
                    throw new Exception('Found empty parameter name');
                }

                $levelParts[$level][] = array('parameter', $matches['name'], isset($matches['delimiters']) ? $matches['delimiters'] : null);

                $currentPos += strlen($matches[0]);
            } elseif ($matches['token'] === '{') {
                if (!preg_match('(\G(?P<literal>[^}]+)\})', $def, $matches, 0, $currentPos)) {
                    throw new Exception('Translated literal missing closing bracket');
                }

                $currentPos += strlen($matches[0]);

                $levelParts[$level][] = array('translated-literal', $matches['literal']);
            } elseif ($matches['token'] === '[') {
                $levelParts[$level][] = array('optional', array());
                $levelParts[$level + 1] = &$levelParts[$level][count($levelParts[$level]) - 1][1];

                $level++;
            } elseif ($matches['token'] === ']') {
                unset($levelParts[$level]);
                $level--;

                if ($level < 0) {
                    throw new Exception('Found closing bracket without matching opening bracket');
                }
            } else {
                break;
            }
        }

        if ($level > 0) {
            throw new Exception('Found unbalanced brackets');
        }

        return $parts;
    }

    /**
     * Build the matching regex from parsed parts.
     *
     * @param  array   $parts
     * @param  array   $params
     * @param  int $groupIndex
     * @return string
     */
    protected function buildRegex(array $parts, array $params, &$groupIndex = 1)
    {
        $regex = '';

        foreach ($parts as $part) {
            switch ($part[0]) {
                case 'literal':
                    $regex .= preg_quote($part[1]);
                    break;

                case 'parameter':
                    $groupName = '?P<param' . $groupIndex . '>';

                    if (isset($params[$part[1]])) {
                        $regex .= '(' . $groupName . $params[$part[1]] . ')';
                    } elseif ($part[2] === null) {
                        $regex .= '(' . $groupName . '[^/]+)';
                    } else {
                        $regex .= '(' . $groupName . '[^' . $part[2] . ']+)';
                    }

                    $this->paramMap['param' . $groupIndex++] = $part[1];
                    break;

                case 'optional':
                    $regex .= '(?:' . $this->buildRegex($part[1], $params, $groupIndex) . ')?';
                    break;
            }
        }

        return $regex;
    }

    /**
     * Build a path.
     *
     * @param  array   $parts
     * @param  array   $mergedParams
     * @param  bool    $isOptional
     * @param  bool    $hasChild
     * @param  array   $options
     * @return string
     * @throws Exception\InvalidArgumentException
     * @throws Exception\RuntimeException
     */
    protected function buildPath(array $parts, array $mergedParams, $isOptional, $hasChild, array $options)
    {

        $path      = '';
        $skip      = true;
        $skippable = false;

        //pre($parts);
        foreach ($parts as $part) {
            switch ($part[0]) {
                case 'literal':
                    $path .= $part[1];
                    break;

                case 'parameter':
                    $skippable = true;

                    if (!isset($mergedParams[$part[1]])) {
                        if (!$isOptional || $hasChild) {
                            throw new Exception(sprintf('Missing parameter "%s"', $part[1]));
                        }

                        return '';
                    } else {
                        $mergedParams[$part[1]] = strval($mergedParams[$part[1]]);

                        //echo $part[1] . ': ' .  ( isset($this->defaults[$part[1]]) ? $this->defaults[$part[1]] : 'null' ) . ' === ' . $mergedParams[$part[1]] . ' ? ' . intval(strval(isset($this->defaults[$part[1]]) ? $this->defaults[$part[1]] : null ) === $mergedParams[$part[1]]) ;
                        if( $isOptional && !$hasChild && $mergedParams[$part[1]] === '' ) {
                            $skip = true;
                        }else if ( !$isOptional || $hasChild || !isset($this->defaults[$part[1]]) || (strval($this->defaults[$part[1]]) !== $mergedParams[$part[1]] ) ) {
                            $skip = false;
                        }
                        //echo ' - $skip: ' . $skip . '<br />';
                    }

                    $path .= $this->encode($mergedParams[$part[1]]);

                    $this->assembledParams[] = $part[1];
                    break;

                case 'optional':
                    $skippable    = true;
                    $optionalPart = $this->buildPath($part[1], $mergedParams, true, $hasChild, $options);

                    if ($optionalPart !== '') {
                        $path .= $optionalPart;
                        $skip  = false;
                    }
                    break;
            }
        }

        if ($isOptional && $skippable && $skip) {
            return '';
        }

        return $path;
    }

    /**
     * Match a given request path.
     *
     * @see    \Zend\Mvc\Router\RouteInterface::match()
     * @param  string $path
     * @param  string|null $pathOffset
     * @param  array       $options
     * @return RouteMatch|null
     * @throws Exception\RuntimeException
     */
    public function match($path, $pathOffset = null, array $options = array())
    {
        $regex = $this->regex;

        if ($pathOffset !== null) {
            $result = preg_match('(\G' . $regex . ')', $path, $matches, null, $pathOffset);
        } else {
            $result = preg_match('(^' . $regex . '$)', $path, $matches);
            //echo 'preg_match(\'(^' . $regex . '$)\', '.$path.'): ' . $result . '<br />';
        }

        if ( !$result ) {
            return null;
        }

        $matchedLength = strlen($matches[0]);

        $params        = array();
        foreach ($this->paramMap as $index => $name) {
            if (isset($matches[$index]) && $matches[$index] !== '') {
                $params[$name] = $this->decode($matches[$index]);
            }
        }

        return new RouteMatch(array_merge($this->defaults,  $this->transformParams( $params, true) ), $matchedLength);
    }

    /**
     * Sprawdza czy dany wzorzec pasuje do parametrw z jakich ma byc zbudowana sciezka
     *
     * @param  array       $params
     * @param  array       $params*
     * @return boolean
     * @throws Exception\RuntimeException
     */
    public function matchParams( &$params, $options = array() )
    {
        if ( !is_array($options) ){
            $options = array(
                'strict' => $options
            );
        }

        $this->debug('route: ' . $this->route );

        $options = array_merge( array(
            'strict' => true,
            'escape' => false
        ), $options );

        $match = true; $in_parts = array();
        foreach ( $this->parts as $part ) {
            if( is_string($part[1]) ){
                $this->debug($part[1] .' => ' . $part[0]);
            }


            if ( $part[0] === 'parameter' ) {
                if ( isset( $params[$part[1]] ) ) {
                    $in_parts[] = $part[1];

                    if ( $options['strict'] && isset($this->params[$part[1]]) ) {
                        if ( !preg_match( $this->_paramRegex($this-> params[$part[1]]), $params[$part[1]]) ) {
                            $match = false;
                            $this->debug('no match: ' . $part[1] . ': '.$params[$part[1]].' => '.$this-> params[$part[1]]);
                            break;
                        }
                    } else if ( empty($params[$part[1]]) ) {
                        $this->debug('empty: ' . $part[1]);
                        $match = false;
                        break;
                    }
                } else if( isset($this->defaults[$part[1]]) ) {
                    continue;
                } else {
                    $this->debug('empty: ' . $part[1]);
                    $match = false;
                    break;
                }
            }
        }

        if ( $match ) {
            foreach( $this->params as $k => $v ) {
                if ( !$this->isOptional($k) ){
                    if( isset($params[$k]) ){
                        if ( !in_array($k, $in_parts) && !preg_match( $this->_paramRegex($v), $params[$k]) ){
                            $match = false;
                            $this->debug( $k . ' =>  ' . $v );
                            break;
                        }
                    } else if ( !isset($this->defaults[$k]) ){
                        $match = false;
                        $this->debug( $k . ' =>  ' . $v . ' is not set' );
                    }
                }
            }
        }

        if ( $options['escape'] ) {

        }

        return $match;
    }

    /**
     * Funkcja przeksztacajca parametry wedug tablicy transformacji te ktre s
     * przekazywane do adresu URL i ponbiej odbierane, tablica transformacji moze zwaierac
     * tablice asocjacyjn przeksztacajc parametry ktre maja zosta przekazane do URL'a
     * i poniej przy odebraniu konwertujaca je spowrotem na takie ktroe s wykorzystywane w
     * skrypcie
     *
     * @param array $params
     * @return array
     */
    function transformParams( $params, $reverse = false ){

        if ( !empty($this->trans) ) {
            $trans = $this->trans;
            if ( $reverse ) {
                foreach( $trans as $k => $v ){
                    if( is_array($v) ){
                        $trans[$k] = array_flip($v);
                    }
                }
            }

            foreach( $params as $k => $v ) {
                if ( isset($trans[$k]) ) {
                    if ( is_array($trans[$k]) ) { //tablica przeksztacenia
                        if ( isset($trans[$k][$v]) ){
                            $params[$k] = $trans[$k][$v];
                        }
                    } else if( is_string($trans[$k]) ) { //funcja
                        $params[$k] = call_user_func ( $trans[$k], $v, $reverse );
                    }
                }
            }
        }

        return $params;
    }

    protected static function debug( $msg ){
        return;
        echo $msg . '<br />' . "\n";
    }

    protected static function _paramRegex( $regex, $subject = null ){
        if ( strpos($regex, '^') === false && strpos($regex, '$') === false ) {
            $regex = '/^' . $regex . '$/';
        } else {
            $regex = '/' . $regex . '/';
        }

        if ( !is_null($subject) ){
            return preg_match($regex, $subject);
        }

        return $regex;
    }

    protected function isOptional( $param, $parts = null )
    {
        if ( is_null($parts) ) {
            $depth = false;
            $parts = $this->parts;
        } else {
            $depth = true;
        }

        foreach ( $parts as $part ) {
            switch ( $part[0] ) {
                case 'parameter':
                    if ( $part[1] == $param ){
                        return ($depth ? true : false);
                    }
                break;

                case 'optional':
                    $isOptional = $this->isOptional( $param, $part[1] );

                    if ( !is_null($isOptional) ) {
                        return $isOptional;
                    }
                break;
            }
        }

        return $depth ? null : false;
    }

    /**
     * Filtruje parametry i usuwa te ktre zostay dopasowane do wzorca
     *
     * @param  array   $params
     * @return array()

     */
    public function filterParams( $params )
    {
        if ( !is_array($params) ) {
            return array();
        }

        foreach( $params as $k => $v ){
            if ( is_null($v) || $v === '' ){
                unset($params[$k]);
            }
        }

        foreach ( $this->paramMap as $p  ) {
            if ( is_array($params) && isset($params[$p]) ) {
                unset($params[$p]);
            }
        }

        foreach ( $this->defaults as $p => $v ) {
            $v = strval($v);

            if ( is_array($params) && isset($params[$p]) && is_scalar($params[$p]) && (strval($params[$p]) === $v) ) {
                unset($params[$p]);
            }
        }

        return is_array($params) ? $params : array();
    }



    /**
     * Assemble the route.
     *
     * @see    \Zend\Mvc\Router\RouteInterface::assemble()
     * @param  array $params
     * @param  array $options
     * @return mixed
     */
    public function assemble(array $params = array(), array $options = array())
    {
        $this->assembledParams = array();

//        echo '<pre >';
//        echo 'buildPath(' . print_r($this->parts,true) .', '.print_r(array_merge($this->defaults, $params),true).' )';
//        echo '</pre>';

        $params = $this->transformParams($params);

        $url = $this->buildPath(
            $this->parts,
            array_merge($this->defaults, $params),
            false,
            (isset($options['has_child']) ? $options['has_child'] : false),
            $options
        );

        return $url;
    }

    /**
     * getAssembledParams(): defined by RouteInterface interface.
     *
     * @see    RouteInterface::getAssembledParams
     * @return array
     */
    public function getAssembledParams()
    {
        return $this->assembledParams;
    }

    /**
     * Encode a path segment.
     *
     * @param  string $value
     * @return string
     */
    protected static function encode($value)
    {
        if (!isset(self::$cacheEncode[$value])) {
            self::$cacheEncode[$value] = rawurlencode($value);
            self::$cacheEncode[$value] = strtr(self::$cacheEncode[$value], self::$urlencodeCorrectionMap);
        }
        return self::$cacheEncode[$value];
    }

    /**
     * Decode a path segment.
     *
     * @param  string $value
     * @return string
     */
    protected static function decode($value)
    {
        return rawurldecode($value);
    }
}
